@@ -1,5 +1,5 @@
 /* GENIUS Calculator
- * Copyright (C) 1997-2004 George Lebl
+ * Copyright (C) 1997-2009 George Lebl
  *
  * Author: George Lebl
  *
@@ -88,22 +88,24 @@ gel_compile_node(GelETree *t,GString *gs)
 	GelETree *ali;
 	g_string_append_printf (gs, ";%d", t->type);
 	switch(t->type) {
-	case NULL_NODE:
+	case GEL_NULL_NODE:
 		break;
-	case VALUE_NODE:
+	case GEL_VALUE_NODE:
 		s = mpw_getstring(t->val.value,0,FALSE,FALSE,FALSE,GEL_OUTPUT_NORMAL,10, TRUE);
 		g_string_append_c(gs,';');
 		g_string_append(gs,s);
 		g_free(s);
 		break;
-	case MATRIX_NODE:
+	case GEL_MATRIX_NODE:
 		g_string_append_printf (gs, ";%dx%d;%d",
 					gel_matrixw_width (t->mat.matrix),
 					gel_matrixw_height (t->mat.matrix),
 					t->mat.quoted);
+		/* Note that compilation goes in opposite to usual order for genius,
+		 * columnwise and not rowwise */
 		for(i=0;i<gel_matrixw_width(t->mat.matrix);i++) {
 			for(j=0;j<gel_matrixw_height(t->mat.matrix);j++) {
-				GelETree *tt = gel_matrixw_set_index(t->mat.matrix,i,j);
+				GelETree *tt = gel_matrixw_get_index(t->mat.matrix,i,j);
 				if(!tt) g_string_append(gs,";0");
 				else {
 					g_string_append(gs,";N");
@@ -112,7 +114,7 @@ gel_compile_node(GelETree *t,GString *gs)
 			}
 		}
 		break;
-	case OPERATOR_NODE:
+	case GEL_OPERATOR_NODE:
 		g_string_append_printf (gs, ";%d;%d",
 					t->op.oper,
 					t->op.nargs);
@@ -120,10 +122,10 @@ gel_compile_node(GelETree *t,GString *gs)
 			gel_compile_node(ali,gs);
 		}
 		break;
-	case IDENTIFIER_NODE:
+	case GEL_IDENTIFIER_NODE:
 		g_string_append_printf (gs, ";%s", t->id.id->token);
 		break;
-	case STRING_NODE:
+	case GEL_STRING_NODE:
 		if(*t->str.str) {
 			g_string_append_c(gs,';');
 			append_string(gs,t->str.str);
@@ -131,10 +133,12 @@ gel_compile_node(GelETree *t,GString *gs)
 			g_string_append(gs,";E");
 		}
 		break;
-	case FUNCTION_NODE:
+	case GEL_FUNCTION_NODE:
 		g_assert(t->func.func->type==GEL_USER_FUNC);
 		/*g_assert(t->func.func->id==NULL);*/
-		g_string_append_printf (gs, ";%d;%d;%d;%d",
+		g_string_append_printf (gs, ";%s;%s;%d;%d;%d;%d",
+					t->func.func->id ? t->func.func->id->token : "*",
+					t->func.func->symbolic_id ? t->func.func->symbolic_id->token : "*",
 					t->func.func->nargs,
 					t->func.func->vararg,
 					t->func.func->propagate_mod,
@@ -145,7 +149,7 @@ gel_compile_node(GelETree *t,GString *gs)
 		}
 		gel_compile_node(t->func.func->data.user,gs);
 		break;
-	case COMPARISON_NODE:
+	case GEL_COMPARISON_NODE:
 		g_string_append_printf (gs, ";%d", t->comp.nargs);
 		for(li=t->comp.comp;li;li=g_slist_next(li)) {
 			int oper = GPOINTER_TO_INT(li->data);
@@ -155,7 +159,7 @@ gel_compile_node(GelETree *t,GString *gs)
 			gel_compile_node(ali,gs);
 		}
 		break;
-	case BOOL_NODE:
+	case GEL_BOOL_NODE:
 		g_string_append_printf (gs, ";%c", 
 					t->bool_.bool_ ? 't' : 'f');
 		break;
@@ -183,6 +187,7 @@ static GelETree *
 gel_decompile_node(char **ptrptr)
 {
 	GelETree *n;
+	const char *id, *symbolic_id;
 	char *p;
 	int type = -1;
 	int nargs = -1;
@@ -206,15 +211,15 @@ gel_decompile_node(char **ptrptr)
 	if G_UNLIKELY (type==-1) return NULL;
 
 	switch(type) {
-	case NULL_NODE:
+	case GEL_NULL_NODE:
 		return gel_makenum_null();
-	case VALUE_NODE:
+	case GEL_VALUE_NODE:
 		p = strtok_r(NULL,";", ptrptr);
 		if G_UNLIKELY (!p) return NULL;
 		mpw_init(tmp);
 		mpw_set_str(tmp,p,10);
 		return gel_makenum_use(tmp);
-	case MATRIX_NODE:
+	case GEL_MATRIX_NODE:
 		p = strtok_r(NULL,";", ptrptr);
 		if G_UNLIKELY (!p) return NULL;
 		h = w = -1;
@@ -230,6 +235,8 @@ gel_decompile_node(char **ptrptr)
 
 		m = gel_matrixw_new();
 		gel_matrixw_set_size(m,w,h);
+		/* Note that compilation goes in opposite to usual order for genius,
+		 * columnwise and not rowwise */
 		for(i=0;i<w;i++) {
 			for(j=0;j<h;j++) {
 				p = strtok_r (NULL,";", ptrptr);
@@ -250,12 +257,12 @@ gel_decompile_node(char **ptrptr)
 				}
 			}
 		}
-		GET_NEW_NODE(n);
-		n->type = MATRIX_NODE;
+		GEL_GET_NEW_NODE(n);
+		n->type = GEL_MATRIX_NODE;
 		n->mat.matrix = m;
 		n->mat.quoted = quote;
 		return n;
-	case OPERATOR_NODE:
+	case GEL_OPERATOR_NODE:
 		p = strtok_r (NULL,";", ptrptr);
 		if G_UNLIKELY (!p) return NULL;
 		oper = -1;
@@ -286,20 +293,20 @@ gel_decompile_node(char **ptrptr)
 			li->any.next = NULL;
 		}
 
-		GET_NEW_NODE(n);
-		n->type = OPERATOR_NODE;
+		GEL_GET_NEW_NODE(n);
+		n->type = GEL_OPERATOR_NODE;
 		n->op.args = args;
 		n->op.nargs = nargs;
 		n->op.oper = oper;
 		return n;
-	case IDENTIFIER_NODE:
+	case GEL_IDENTIFIER_NODE:
 		p = strtok_r (NULL,";", ptrptr);
 		if G_UNLIKELY (!p) return NULL;
-		GET_NEW_NODE(n);
-		n->type = IDENTIFIER_NODE;
+		GEL_GET_NEW_NODE(n);
+		n->type = GEL_IDENTIFIER_NODE;
 		n->id.id = d_intern(p);
 		return n;
-	case STRING_NODE:
+	case GEL_STRING_NODE:
 		p = strtok_r (NULL, ";", ptrptr);
 		if G_UNLIKELY (p == NULL)
 			return NULL;
@@ -314,7 +321,21 @@ gel_decompile_node(char **ptrptr)
 			g_free (p);
 		}
 		return n;
-	case FUNCTION_NODE:
+	case GEL_FUNCTION_NODE:
+		p = strtok_r (NULL,";", ptrptr);
+		if G_UNLIKELY (!p) return NULL;
+		if (strcmp (p, "*") == 0)
+			id = NULL;
+		else
+			id = p;
+
+		p = strtok_r (NULL,";", ptrptr);
+		if G_UNLIKELY (!p) return NULL;
+		if (strcmp (p, "*") == 0)
+			symbolic_id = NULL;
+		else
+			symbolic_id = p;
+
 		p = strtok_r (NULL,";", ptrptr);
 		if G_UNLIKELY (!p) return NULL;
 		nargs = -1;
@@ -352,17 +373,19 @@ gel_decompile_node(char **ptrptr)
 			return NULL;
 		}
 
-		func = d_makeufunc(NULL,n,oli,nargs, NULL);
+		/* Note that d_intern handles NULL properly */
+		func = d_makeufunc (d_intern (id), n, oli, nargs, NULL);
+		func->symbolic_id = d_intern (symbolic_id);
 		func->context = -1;
 		func->vararg = vararg ? 1 : 0;
 		func->propagate_mod = propagate_mod ? 1 : 0;
 		func->no_mod_all_args = no_mod_all_args ? 1 : 0;
 
-		GET_NEW_NODE(n);
-		n->type = FUNCTION_NODE;
+		GEL_GET_NEW_NODE(n);
+		n->type = GEL_FUNCTION_NODE;
 		n->func.func = func;
 		return n;
-	case COMPARISON_NODE:
+	case GEL_COMPARISON_NODE:
 		p = strtok_r (NULL,";", ptrptr);
 		if G_UNLIKELY (!p) return NULL;
 		sscanf(p,"%d",&nargs);
@@ -404,13 +427,13 @@ gel_decompile_node(char **ptrptr)
 			li->any.next = NULL;
 		}
 
-		GET_NEW_NODE(n);
-		n->type = COMPARISON_NODE;
+		GEL_GET_NEW_NODE(n);
+		n->type = GEL_COMPARISON_NODE;
 		n->comp.args = args;
 		n->comp.nargs = nargs;
 		n->comp.comp = oli;
 		return n;
-	case BOOL_NODE:
+	case GEL_BOOL_NODE:
 		p = strtok_r (NULL, ";", ptrptr);
 		if G_UNLIKELY (p == NULL)
 			return NULL;
@@ -443,6 +466,7 @@ gel_decompile_tree (char *s)
 	
 	t = gel_decompile_node (&ptrptr);
 	if G_UNLIKELY (t == NULL) {
+		printf ("FOO!\n");
 		gel_errorout (_("Bad tree record when decompiling"));
 		return NULL;
 	}
